from .context import session

# Import all of the client classes for type annotations.
# We have to do this because we lazy-import the client definitions
# inside each function to reduce runtime dependencies.
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from openstack.connection import Connection
    from blazarclient.client import Client as BlazarClient
    from glanceclient.client import Client as GlanceClient
    from gnocchiclient.v1.client import Client as GnocchiClient
    from manilaclient.client import Client as ManilaClient
    from neutronclient.v2_0.client import Client as NeutronClient
    from novaclient.client import Client as NovaClient
    from ironicclient import client as IronicClient
    from keystoneclient.v3.client import Client as KeystoneClient
    from zunclient.client import Client as ZunClient


session_factory = session

NOVA_API_VERSION = "2.10"
ZUN_API_VERSION = "1.41"


def connection(session=None) -> "Connection":
    """Get connection context for OpenStack SDK.

    The returned :class:`openstack.connection.Connection` object has
    several proxy modules attached for each service provided by the cloud.

    .. note::
       For the most part, it is more straightforward to use clients specific
       to the service you are targeting. However, some of the proxy modules
       are useful for operations that span a few services, such as assigning
       a Floating IP to a server instance.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new connection proxy.
    """
    from openstack.config import cloud_region
    from openstack.connection import Connection

    sess = session or session_factory()
    if hasattr(sess, "session"):
        # Handle Adapters, which have a nested Session
        sess = sess.session
    cloud_config = cloud_region.from_session(sess)
    return Connection(
        config=cloud_config,
        compute_api_version=NOVA_API_VERSION,
    )


def blazar(session=None) -> "BlazarClient":
    """Get a preconfigured client for Blazar, the reservation service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Blazar client.
    """
    from blazarclient.client import Client as BlazarClient

    return BlazarClient(
        "1", service_type="reservation", session=(session or session_factory())
    )


def glance(session=None) -> "GlanceClient":
    """Get a preconfigured client for Glance, the image service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Glance client.
    """
    from glanceclient.client import Client as GlanceClient

    return GlanceClient("2", session=(session or session_factory()))


def gnocchi(session=None) -> "GnocchiClient":
    """Get a preconfigured client for Gnocchi, the metrics service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Gnocchi client.
    """
    from gnocchiclient.v1.client import Client as GnocchiClient

    sess = session or session_factory()
    session_options = dict(auth=sess.session.auth)
    adapter_options = dict(interface=sess.interface, region_name=sess.region_name)
    return GnocchiClient(
        adapter_options=adapter_options, session_options=session_options
    )


def manila(session=None) -> "ManilaClient":
    """Get a preconfigured client for Manila, the share service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Manila client.
    """
    from manilaclient.client import Client as ManilaClient

    return ManilaClient("2", session=(session or session_factory()))


def neutron(session=None) -> "NeutronClient":
    """Get a preconfigured client for Neutron, the networking service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Neutron client.
    """
    from neutronclient.v2_0.client import Client as NeutronClient

    return NeutronClient(session=(session or session_factory()))


def nova(session=None) -> "NovaClient":
    """Get a preconfigured client for Nova, the compute service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Nova client.
    """
    from novaclient.client import Client as NovaClient

    return NovaClient(NOVA_API_VERSION, session=(session or session_factory()))


def ironic(session=None) -> "IronicClient":
    """Get a preconfigured client for Ironic, the bare metal service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Ironic client.
    """
    from ironicclient import client as IronicClient

    return IronicClient.get_client(
        "1",
        session=(session or session_factory()),
        region_name=getattr(session, "region_name", None),
        # Ironic client defaults to 1.9 currently,
        # "latest" will be latest the API supports
        os_ironic_api_version="latest",
    )


def keystone(session=None) -> "KeystoneClient":
    """Get a preconfigured client for Keystone, the authentication service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Keystone client.
    """
    from keystoneclient.v3.client import Client as KeystoneClient

    sess = session or session_factory()
    # We have to set interface/region_name also on the Keystone client, as it
    # does not smartly inherit the value sent in on a KSA Adapter instance.
    return KeystoneClient(
        session=sess,
        interface=getattr(sess, "interface", None),
        region_name=getattr(sess, "region_name", None),
    )


def zun(session=None) -> "ZunClient":
    from zunclient.client import Client as ZunClient

    return ZunClient(ZUN_API_VERSION, session=(session or session_factory()))



from chi.ssh import Remote

LOCALHOST = "127.0.0.1"


def _context_get_keypair_private_key(key):
    return "fake_key"


def test_get(mocker):
    mocker.patch("chi.context.get",
                 side_effect=_context_get_keypair_private_key)
    r = Remote(ip=LOCALHOST)
    assert r.host == LOCALHOST
    assert r.user == "cc"


def test_get_from_server(mocker):
    class FakeServer:
        def __init__(self):
            self.ip = LOCALHOST

    mocker.patch("chi.context.get",
                 side_effect=_context_get_keypair_private_key)
    r = Remote(server=FakeServer())
    assert r.host == LOCALHOST
    assert r.user == "cc"


from .clients import manila

from manilaclient.exceptions import NotFound

__all__ = [
    'create_share',
    'delete_share',
    'extend_share',
    'get_access_rules',
    'get_share',
    'get_share_id',
    'list_shares',
    'shrink_share',
]


def _get_default_share_type_id():
    # we only support one share type - cephfsnfstype
    share_types = manila().share_types.list()
    if not share_types:
        raise ValueError("No share types found")
    elif len(share_types) > 1:
        raise ValueError("Multiple share types found")
    return share_types[0].id


def create_share(size, name=None, description=None, metadata=None,
                 is_public=False):
    """Create a share.

    Args:
        size (int): size in GiB.
        name (str): name of new share.
        description (str): description of a share.
        is_public (bool): whether to set share as public or not.

    Returns:
        The created share.
    """
    share = manila().shares.create(
        share_proto="NFS",
        size=size,
        name=name,
        description=description,
        metadata=metadata,
        share_type=_get_default_share_type_id(),
        is_public=is_public
    )
    return share


def delete_share(share):
    """Delete a share.

    Args:
        share: either share object or text with its ID.
    """
    manila().shares.delete(share)


def extend_share(share, new_size):
    """Extend the size of the specific share.

    Args:
        share: either share object or text with its ID.
        new_size: desired size to extend share to.
    """
    manila().shares.extend(share, new_size)


def get_access_rules(share):
    """Get access list to a share.

    Args:
        share: either share object or text with its ID.

    Returns:
        A list of access rules.
    """
    return manila().shares.access_list(share)


def get_share(ref):
    """Get a share by its ID or name.

    Args:
        ref (str): The ID or name of the share.

    Returns:
        The share matching the ID or name.

    Raises:
        NotFound: If the share could not be found.
    """
    try:
        return manila().shares.get(ref)
    except NotFound:
        return manila().shares.get(get_share_id(ref))


def get_share_id(name):
    """Look up a share's ID from its name.

    Args:
        name (str): The name of the share.

    Returns:
        The ID of the found share.

    Raises:
        ValueError: If the share could not be found, or if multiple shares
            matched the name.
    """
    shares = list(manila().shares.list(filters={'name': name}))
    if not shares:
        raise ValueError(f'No shares found matching name "{name}"')
    elif len(shares) > 1:
        raise ValueError(f'Multiple shares found matching name "{name}"')
    return shares[0].id


def list_shares():
    """List all shares under the current project.

    Returns:
        All shares associated with the current project.
    """
    return list(manila().shares.list())


def shrink_share(share, new_size):
    """Shrink the size of the specific share.

    Args:
        share: either share object or text with its ID.
        new_size: desired size to shrink share to.
    """
    manila().shares.shrink(share, new_size)



# Copyright 2021 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def example_create_share():
    """Create a share.

    <div class="alert alert-info">

    **Functions used in this example:**

    * [create_share](../modules/share.html#chi.share.create_share)

    </div>

    """
    from chi.share import create_share

    share_name = "my_share"
    share = create_share(size=1, name=share_name)


def test_example_create_share(mocker):
    manila = mocker.patch("chi.share.manila")()

    def _get_default_share_type_id():
        return "fakesharetypeid"

    mocker.patch("chi.share._get_default_share_type_id",
                 side_effect=_get_default_share_type_id)

    example_create_share()

    manila.shares.create.assert_called_once_with(
        share_proto='NFS',
        size=1,
        name="my_share",
        description=None,
        metadata=None,
        share_type=_get_default_share_type_id(),
        is_public=False
    )
